#!/usr/bin/python

import os
import sys
import re
import trace_reader
import trace_event
from execution_state import ExecutionState
from trace import Trace


class Analyze:

  def __init__(self):
    line_number = None

  def GetStatsDict(self, exec_state):
    stats = {}
    agg = exec_state.aggregates
    for attr in agg.__dict__:
      if attr.startswith('_'):
        continue  # skip hidden variables
      stat_name = attr.replace('_', '-')
      stats[stat_name] = getattr(agg, attr)
    return stats

  def OutputStats(self, exec_state, out_dir, full_name):
    # output regular stats
    dir_path = os.path.join(out_dir, 'stats')
    if not os.path.exists(dir_path):
      os.mkdir(dir_path)
    out_path = os.path.join(dir_path,
                            '%s.txt' % full_name)
    f = open(out_path, 'w')
    stats = self.GetStatsDict(exec_state)
    for key in stats:
      f.write('%s: %s\n' % (key, stats[key]))
    f.close()

    # output stats from listeners
    for listener in exec_state.listeners:
      listner_name = listener.GetName()
      dir_path = os.path.join(out_dir, listner_name)
      if not os.path.exists(dir_path):
        os.mkdir(dir_path)
      listener.OutputToDir(dir_path, full_name)
      

  def HandleEvent(self, exec_state, event):
    generates_fd = event.generates_fd
    probe_type = event.a_probe_type
    if generates_fd and probe_type == 'return':
      event.execute(exec_state)
    if not generates_fd and probe_type == 'entry':
      event.execute(exec_state)

  def GetAppName(self, trace_path):
    trace_file = os.path.basename(trace_path)
    m = re.match(r'([^/]*)\_[^/]*\.tar\.gz$', trace_file)
    if not m:
      msg = ('Cannot determine app name from trace file'
             '%s.  Please use format AppName_TaskNameSize')
      raise Exception(msg)
    app_name = m.group(1)
    return app_name
    

  def _ProcessTrace(self, trace_path, show_progress):
    print 'Processing trace: %s' % trace_path
    self.line_number = None

    # load the trace from the .tar.gz
    full_name = os.path.basename(trace_path).replace('.tar.gz', '')
    app_name = self.GetAppName(trace_path)
    trace = Trace(app=app_name, file_path=trace_path)
    exec_state = ExecutionState()

    # process snapshot before the trace events
    for inode in trace.trace_reader.snapshot:
      exec_state.AddInode(inode)

    # process all events in the trace
    self.line_number = 1
    last_percent = 0
    first_time = True
    start_timestamp = None
    end_startup_timestamp = None

    for event in trace:
      # print progress info
      if show_progress:
        percent = trace.trace_reader.Percent()
        if percent != last_percent:
          last_percent = percent
          if percent % 10 == 0:
            print 'Progress %d%%' % percent

      # record trace start time
      if first_time:
        first_time = False
        start_timestamp = event.a_time

      if isinstance(event, trace_event.Event_controller):
        if event.a_msg == 'END STARTUP':
          exec_state.FinishStartup()
          end_startup_timestamp = event.a_time
      else:
        self.HandleEvent(exec_state, event)
      self.line_number += 1
    exec_state.FinishTrace()
    exec_state.AddCPUStats(start_timestamp,
                           end_startup_timestamp,
                           trace.trace_reader.last_timestamp,
                           trace.trace_reader.cpu_values)

    # save resulting stats to a file
    out_dir = os.path.dirname(os.path.abspath(trace_path))
    self.OutputStats(exec_state, out_dir, full_name)

  def ProcessTrace(self, trace_path, catch_exceptions, show_progress):
    if not catch_exceptions:
      self._ProcessTrace(trace_path, show_progress)
      return
    try:
      self._ProcessTrace(trace_path, show_progress)
    except Exception as e:
      if self.line_number:
        print 'FAILED on trace event %d' % self.line_number
      else:
        print 'FAILED'

  def main(self):
    # grab startup arguments
    path = None
    catch_exceptions = False
    show_progress = False
    i = 1
    while i < len(sys.argv):
      if sys.argv[i] == '-c':
        catch_exceptions = True
      elif sys.argv[i] == '-p':
        show_progress = True
      else:
        path = sys.argv[i]
      i += 1
    if path == None:
      raise Exception('usage: analyze.py <trace_file.tar.gz|trace_directory> [-c] [-p]')

    # process trace or directory of traces
    if os.path.isdir(path):
      files = os.listdir(path)
      tars = []
      for x in files:
        if x.endswith('.tar.gz'):
          tars.append('%s/%s' % (path, x))
      if len(tars) == 0:
        raise Exception('No .tar.gz traces files could '
                        'be found in %s' % trace_path)
      for x in tars:
        self.ProcessTrace(x, catch_exceptions, show_progress)
    else:
      self.ProcessTrace(path, catch_exceptions, show_progress)


if __name__ == '__main__': 
  analyze = Analyze()
  analyze.main()




