import re
import trace_reader
import close_listeners

class ExecutionState(object):
  def __init__(self):
    # aggregates used to generate stats
    self.aggregates = Aggregates()
    # key = file descriptor integer, val = Fd object
    self.fds = {}
    # key = inode identifier, val = Inode object
    self.inodes = {}
    # key = tid, val = Thread object
    self.threads = {}

    self.listeners = []
    self.listeners.append(close_listeners.FileSizeCDF())
    
    # ignore stdin, stdout, stderr
    self.AddDummyFd(0)
    self.AddDummyFd(1)
    self.AddDummyFd(2)

  def FinishStartup(self):
    self.aggregates.Reset()
    for fd_number in self.fds:
      self.fds[fd_number].Reset()
    for tid in self.threads:
      self.threads[tid].Reset()
    for inode in self.inodes:
      self.inodes[inode].Reset()

  def FinishTrace(self):
    # sometimes programs don't automatically close their FDs, so we need to do
    # it for them so that the aggregate stats will be updated
    self.CloseAllFds()

    # count how many threads do I/O
    agg = self.aggregates
    count = len([t for t in 
                 self.threads.values()
                 if t.TotalIO() > 0])
    agg.io_thread_count = count

    count = len([t for t in 
                 self.threads.values()
                 if t.read_bytes > 0
                 and t.write_bytes == 0])
    agg.threads_read_count = count

    count = len([t for t in 
                 self.threads.values()
                 if t.write_bytes > 0
                 and t.read_bytes == 0])
    agg.threads_write_count = count

    count = len([t for t in 
                 self.threads.values()
                 if t.write_bytes > 0
                 and t.read_bytes > 0])
    agg.threads_both_count = count

    count = len([t for t in 
                 self.threads.values()
                 if t.write_bytes == 0
                 and t.read_bytes > 0
                 and t.fsync_count > 0])
    agg.threads_read_and_fsync_count = count

    count = len([t for t in 
                 self.threads.values()
                 if t.write_bytes > 0
                 and t.read_bytes == 0
                 and t.fsync_count > 0])
    agg.threads_write_and_fsync_count = count

    count = len([t for t in 
                 self.threads.values()
                 if t.write_bytes > 0
                 and t.read_bytes > 0
                 and t.fsync_count > 0])
    agg.threads_both_and_fsync_count = count

    # see how much I/O the top 30 threads perform
    unsorted_threads = self.threads.values()
    sorted_threads = sorted(unsorted_threads, key=lambda thread: -thread.TotalIO())
    for i in range(30):
      bytes = 0
      if i < len(sorted_threads):
        bytes = sorted_threads[i].TotalIO()
      setattr(self.aggregates, 'thread%d_io' % (i + 1), bytes)

    # count I/O for each file type
    for inode in self.inodes.values():
      cat = inode.category
      if cat != None:
        # update bytes for this category
        attr = 'total_bytes_category_%s' % cat
        value = getattr(agg, attr)
        value += inode.TotalIO()
        setattr(agg, attr, value)
        # update accesses for this category
        attr = 'total_accesses_category_%s' % cat
        value = getattr(agg, attr)
        value += inode.accesses
        setattr(agg, attr, value)
      # update fsync bytes for each cause
      for fsync_cause in inode.fsync_cause_bytes:
        attr = 'fsync_bytes_%s' % fsync_cause
        value = getattr(agg, attr)
        value += inode.fsync_cause_bytes[fsync_cause]
        setattr(agg, attr, value)
      # update rename counts for each cause
      for rename_cause in inode.rename_cause_counts:
        attr = 'rename_count_cause_%s' % rename_cause
        value = getattr(agg, attr)
        value += inode.rename_cause_counts[rename_cause]
        setattr(agg, attr, value)

  def AddCPUStats(self,
                  start_timestamp,
                  end_startup_timestamp,
                  end_timestamp,
                  cpu_values):
    agg = self.aggregates
    if end_startup_timestamp is None:
      # this was a startup experiment, so we want to
      # count all the CPU seconds
      for x in cpu_values:
        agg.cpu_seconds += x
    else:
      # seconds between END STARTUP and last timestamp
      phase2 = float(end_timestamp - end_startup_timestamp) / 1e9

      # we need to determine the maximinum number of cpu_values 
      # frames could occur after END STARTUP in
      # the event trace.  We round up the number
      # of seconds up, then add one since two frames
      # could overlap one second.
      max_frames = int(phase2 + 0.5) + 1

      # get an upper bound on the cpu seconds
      i = len(cpu_values) - max_frames
      while i < len(cpu_values):
        agg.cpu_seconds += cpu_values[i]
        i += 1

  def OpenFile(self,
               timestamp,
               fd_number,
               inode_number,
               path,
               flags):
    # make sure we have the inode in our hash
    if inode_number not in self.inodes:
      if not 'O_CREAT' in flags:
        if int(inode_number.rstrip('rsrc')) in self.inodes:
          # we don't need O_CREAT for rsrc forks
          new_inode = Inode(inode_number, path, 0)
          self.AddInode(new_inode)
        else:
          msg = ('Inode %s does not exist, and O_CREAT '
                 'was not used' % inode_number)
          raise trace_reader.AnalyzeException(msg)
      else:
        new_inode = Inode(inode_number, path, 0)
        self.AddInode(new_inode)
    
    # grab the inode
    inode = self.inodes[inode_number]

    # check if the file was truncated
    if 'O_TRUNC' in flags:
      inode.size = 0

    # add the FD to our hash
    fd_object = FileFd(timestamp=timestamp,inode=inode)

    # check if we're in append only mode
    if 'O_APPEND' in flags:
      fd_object.append_only = True

    self.AddFd(fd_number=fd_number, fd_object=fd_object)

  def CloseAllFds(self):
    fd_list = [fd_number for fd_number in self.fds]
    for fd_number in fd_list:
      self.CloseFd(fd_number)

  def AddFd(self, fd_number, fd_object):
    if fd_number < 0:
      raise Exception('negative fd')

    if fd_number in self.fds:
      raise trace_reader.AnalyzeException('FD %d already in use' % fd_number)
    fd_object.exec_state = self
    self.fds[fd_number] = fd_object

  def DupFd(self, old_fd, new_fd):
    if not old_fd in self.fds:
      raise trace_reader.AnalyzeException('FD %d already in use' % old_fd)
    if new_fd in self.fds:
      self.CloseFd(new_fd)
    self.fds[new_fd] = self.fds[old_fd]
    self.fds[new_fd].ref_count += 1

  def CloseFd(self, fd_number):
    if not fd_number in self.fds:
      raise trace_reader.AnalyzeException('FD %d does not exist' % fd_number)
    self.fds[fd_number].Close()
    del self.fds[fd_number]

  def AddInode(self, inode):
    if inode.number in self.inodes:
      raise trace_reader.AnalyzeException('Inode %s already in use' % inode.number)
    self.inodes[inode.number] = inode
    inode.exec_state = self

  def GetInode(self, inode_number):
    return self.inodes[inode_number]

  def AddDummyFd(self, fd_number):
    self.AddFd(fd_number, IgnoreFd())

  def GetFd(self, fd_number):
    return self.fds[fd_number]

  def GetOrCreateThread(self, tid):
    if tid not in self.threads:
      self.threads[tid] = Thread(tid)
    return self.threads[tid]


class Inode(object):
  CAT_PLIST = 'plist'
  CAT_STRINGS = 'extension_strings'
  CAT_SQLITE = 'sqlite'
  CAT_PRODUCTIVITY = 'productivity'
  CAT_MULTIMEDIA = 'multimedia'

  def __init__(self, number, path, size):
    self.number = number
    self.category = None
    # setting the path may override the category
    self.SetPath(path)
    self.size = size
    self.exec_state = None
    self.Reset()

  def Reset(self):
    self.read_bytes = 0
    self.write_bytes = 0
    self.accesses = 0
    # how many bytes have been written this inode
    # prior to a rename or exchangedata?
    self.not_moved_bytes = 0
    # key = fsync cause, 
    # value = bytes fsynced by the cause
    self.fsync_cause_bytes = {}
    # key = rename cause, 
    # value = number of renames
    self.rename_cause_counts = {}

  def TotalIO(self):
    return self.read_bytes + self.write_bytes

  def GetCategoryFromPath(self, path):
    # determine category from path
    multimedia_extensions = ['jpg', 'jpeg', 'tiff',
                             'tif', 'png', 'mov',
                             'm4v', 'mp3', 'aiff']
    productivity_extensions = ['pages', 'key', 'numbers',
                               'xls', 'doc', 'ppt', 'pdf']
    cat = None
    extension = None
    m = re.search(r'\.([^\.]+)$', path)
    if m:
      extension = m.group(1)
      extension = extension.lower()

    if extension in multimedia_extensions:
      cat = Inode.CAT_MULTIMEDIA
    if extension in productivity_extensions:
      cat = Inode.CAT_PRODUCTIVITY
    elif extension == 'plist':
      cat = Inode.CAT_PLIST
    elif extension == 'strings':
      cat = Inode.CAT_STRINGS
    # Note: sqlite causes are determined by looking at
    # the stacks instead of looking at the extension since
    # extensions on sqlite files are not very helpful.
    return cat

  def SetCategory(self, category):
    if category == None:
      return
    elif self.category == None or self.category == category:
      self.category = category
    else:
      raise Exception('We have tried to categorize '
                      '%s as both a %s and a %s '
                      'due to conflicting evidence.'
                      % (self, self.category, category))

  def SetPath(self, path):
    self.path = path
    self.SetCategory(self.GetCategoryFromPath(path))
      
  def Rename(self, stack, new_path, dir_inode_num_1, dir_inode_num_2):
    self.exec_state.aggregates.rename_count += 1
    aggs = self.exec_state.aggregates
    bytes = self.not_moved_bytes
    self.not_moved_bytes = 0
    # should we count the bytes as being in the same
    # directory or moving across directories?
    if dir_inode_num_1 != dir_inode_num_2:
      aggs.rename_bytes_diff_dir += bytes
    else:
      aggs.rename_bytes_same_dir += bytes
    self.SetPath(new_path)

    #rename causes
    cause = self.GetRenameCauseFromStack(stack)
    if not cause in self.rename_cause_counts:
      self.rename_cause_counts[cause] = 0
    self.rename_cause_counts[cause] += 1

  def GetRenameCauseFromStack(self, stack):
    # TODO - this function is very ad hoc for the function
    # calls we saw.  For more general use, this should
    # probably be rewritten at some point.
    cause_map = [
      (r'', r'^CFPreferences', 'preferences_synchronize'),
      (r'', r'NSData.*writeToFile', 'nsdata_writetofile'),
      (r'', r'NSFileManager.*movePath', 'nsfilemanager_movepath'),
      (r'carbon', r'FSRenameUnicode', 'carbon_fsrenameunicode')]

    cause_count = 0
    cause = None
    frames = stack.split('|')
    for frame in frames:
      library, func = frame.split('`')
      func = re.sub(r'\+?0x[0-9a-fA-F]+', r'', func)
      for lib_filter, func_filter, cause_name in cause_map:
        if re.search(lib_filter, library, re.I):
          if re.search(func_filter, func, re.I):
            if cause_name != cause:
              cause = cause_name
              cause_count += 1
    if cause_count > 1:
      return 'multiple'
    elif cause_count == 0:
      return 'unknown'
    else:
      return cause

  def ExchangeData(self, other_inode, new_path, same_dir):
    # exchange data gets called on both inodes, but
    # we only want to increment the count onces, so
    # we just do it if we are the smaller inode
    if self.number < other_inode.number:
      self.exec_state.aggregates.exchangedata_count += 1
    elif self.number == other_inode.number:
      msg = ('What does it mean to do exchangedata on '
             'an inode with itself')
      raise Exception(msg)

    aggs = self.exec_state.aggregates
    bytes = self.not_moved_bytes
    self.not_moved_bytes = 0
    # should we count the bytes as being in the same
    # directory or moving across directories?
    if not same_dir:
      aggs.exchangedata_bytes_diff_dir += bytes
    else:
      aggs.exchangedata_bytes_same_dir += bytes
    self.SetPath(new_path)

  def __str__(self):
    return 'Inode(%s)' % self.number

  
class Fd(object):
  def __init__(self):
    self.ignore = True
    self.exec_state = None
    # if dup is used, multiple integers can refence
    # the same file descriptor
    self.ref_count = 1
  def Reset(self):
    pass
  def TotalIO(self):
    pass
  def Read(self, bytes, offset=None):
    pass
  def Write(self, bytes, offset=None):
    pass
  def Fsync(self):
    pass
  def Ftruncate(self):
    pass
  def Close(self):
    pass

class FileFd(Fd):
  def __init__(self, timestamp, inode):
    super(FileFd, self).__init__()
    self.ignore = False
    self.inode = inode
    self.timestamp = timestamp
    # where the next I/O will occur if there's
    # not an offset specified with the request
    self.offset_pointer = 0
    self.append_only = False
    self.Reset()
    
  def Reset(self):
    self.read_bytes = 0
    self.aio_read_bytes = 0
    self.write_bytes = 0
    # how many times was an I/O request non-sequential?
    self.jumps = 0
    # the end offset of the last read or write
    self.last_io_end = 0
    # used to compute partially sequential stats
    self.longest_run = 0
    self.current_run = 0
    # used to compute fsync stats
    # each fsync_X_count variable counts the number
    # of fsyncs at least X large but smaller than the
    # next category
    self.not_fsynced_bytes = 0
    self.fsync_bytes = 0
    self.fsync_count = 0
    self.fsync_10mb_count = 0
    self.fsync_1mb_count = 0
    self.fsync_64kb_count = 0
    self.fsync_4kb_count = 0
    self.fsync_1b_count = 0
    self.fsync_0b_count = 0
    # how many bytes were preallocated by writing at
    # an offset past the end of the file or calling
    # ftruncate?
    self.pwrite_preallocated_bytes = 0
    self.ftruncate_preallocated_bytes = 0
    self.extension_bytes = 0
    # these are for determining whether a preallocation
    # did anything useful
    self.preallocated_bytes_unnecessary = 0
    self.preallocated_last_time = False
    self.preallocated_start = 0
    self.preallocated_end = 0

  def TotalIO(self):
    return self.read_bytes + self.write_bytes

  def TotalPreallocated(self):
    return (self.pwrite_preallocated_bytes +
            self.ftruncate_preallocated_bytes)

  def IsReadOnly(self):
    return (self.read_bytes > 0 and self.write_bytes == 0)

  def IsWriteOnly(self):
    return (self.read_bytes == 0 and self.write_bytes > 0)

  def Seek(self, offset):
    self.offset_pointer = offset

  def _ReadOrWrite(self, stack, bytes, offset=None):
    # we only update the offset pointer when an offset
    # was not specified
    update_offset_pointer = False
    if offset == None:
      update_offset_pointer = True
      offset = self.offset_pointer

    # did we jump position?
    if offset != self.last_io_end and self.TotalIO() > 0:
      self.current_run = 0
      self.jumps += 1

    # how long has the sequential run of I/O been?
    # is it longer than previous runs?
    self.current_run += bytes
    if self.current_run > self.longest_run:
      self.longest_run = self.current_run

    # update offset pointer if necessary
    if update_offset_pointer:
      self.offset_pointer = offset + bytes
    self.last_io_end = offset + bytes

    # does the stack provide any hints about what
    # type of file we're dealing with?
    if stack.find('sqlite') >= 0:
      self.inode.SetCategory(Inode.CAT_SQLITE)

  def Read(self, tid, stack, bytes, offset=None, aio=False):
    self._ReadOrWrite(stack, bytes, offset)
    # update I/O done per FD, TID, and Inode
    self.read_bytes += bytes
    self.exec_state.GetOrCreateThread(tid).read_bytes += bytes
    self.inode.read_bytes += bytes
    if aio:
      self.aio_read_bytes += bytes

    # sanity check.  we haven't written code to cover the
    # the case of a read on an append only file yet.
    if self.append_only:
      raise Exception('Unsupported: read with O_APPEND')

  def Write(self, tid, stack, bytes, offset=None):
    # see if we should override the offset
    if self.append_only:
      # TODO - self._ReadOrWrite only updates self.offset_point
      # when offset is None, so it potentially wrong to override
      # offset on an append only file if we will also read on
      # that file.  We raise an exception in read if this happens.
      offset = self.inode.size

    self._ReadOrWrite(stack, bytes, offset)
    # update I/O done per FD, TID, and Inode
    self.write_bytes += bytes
    self.exec_state.GetOrCreateThread(tid).write_bytes += bytes
    self.inode.write_bytes += bytes
    # track transactional writes (those that are fsynced or renamed)
    self.not_fsynced_bytes += bytes
    self.inode.not_moved_bytes += bytes

    # We use last_io_end-bytes instead of offset since offset
    # may be None.  This only works if we do this *after*
    # updating last_io_end.
    actual_offset = self.last_io_end - bytes

    # determine if there was a last preallocation that was wasted
    actual_offset = self.last_io_end - bytes
    if self.preallocated_last_time:
      write_start = actual_offset
      write_end = actual_offset + bytes

      if write_start <= self.preallocated_start:
        if write_end >= self.preallocated_end:
          self.preallocated_bytes_unnecessary += (
            self.preallocated_end - self.preallocated_start)

    # determine if there was preallocation this time
    preallocated = actual_offset - self.inode.size
    if preallocated < 0:
      preallocated = 0
    self.pwrite_preallocated_bytes += preallocated
    if preallocated > 0:
      self.preallocated_last_time = True
      self.preallocated_start = self.inode.size
      self.preallocated_end = self.last_io_end
    else:
      self.preallocated_last_time = False

    # was the file extended
    if self.last_io_end > self.inode.size:
      ext = (self.last_io_end - self.inode.size)
      self.extension_bytes += ext
      self.inode.size = self.last_io_end

  def Fsync(self, tid, stack):
    bytes = self.not_fsynced_bytes
    # general fsync stats
    self.not_fsynced_bytes = 0
    self.fsync_count += 1
    self.fsync_bytes += bytes

    # fsync sizes
    if bytes >= 10*1024*1024:
      self.fsync_10mb_count += 1
    elif bytes >= 1024*1024:
      self.fsync_1mb_count += 1
    elif bytes >= 64*1024:
      self.fsync_64kb_count += 1
    elif bytes >= 4*1024:
      self.fsync_4kb_count += 1
    elif bytes >= 1:
      self.fsync_1b_count += 1
    else:
      self.fsync_0b_count += 1

    # fsyncs per thread
    self.exec_state.GetOrCreateThread(tid).fsync_count += 1

    #fsync causes
    cause = self.GetFsyncCauseFromStack(stack)
    if cause != None:
      if not cause in self.inode.fsync_cause_bytes:
        self.inode.fsync_cause_bytes[cause] = 0
      self.inode.fsync_cause_bytes[cause] += bytes

  def GetFsyncCauseFromStack(self, stack):
    cause_map = {'sqlite':'sqlite',
                 'CFPreferences':'preferences',
                 'SFArchiving':'sfarchiving',
                 'writeToFile:atomically':'writetofile_atomically',
                 'FSFlushFork':'fsflushfork',}
    cause_count = 0
    cause = None
    for stack_method in cause_map:
      if stack.find(stack_method) >= 0:
        new_cause = cause_map[stack_method]
        if new_cause != cause:
          cause_count += 1
          cause = new_cause
    if cause_count > 1:
      raise Exception('We have not considered multiple fsync causes.')
    return cause

  def Ftruncate(self, length):
    preallocated = (length - self.inode.size)
    if preallocated > 0:
      self.ftruncate_preallocated_bytes += preallocated
    self.inode.size = length

  def Close(self):
    self.ref_count -= 1

    # this condition is true, there are other numbers
    # that still refer to the same file descriptor, so
    # we should not calculate stats yet
    if self.ref_count > 0:
      return

    if self.TotalIO() > 0:
      self.inode.accesses += 1
    # update aggregates stats based on operations
    # performed on this file descriptor
    agg = self.exec_state.aggregates
    self.UpdateAccessAggregates()
    self.UpdateBytesAggregates()
    self.UpdateFsyncAggregates()
    self.UpdateFileAggregates()
    self.UpdateOtherListeners()

  def UpdateAccessAggregates(self):
    agg = self.exec_state.aggregates
    did_io = (self.TotalIO() > 0)
    did_reads = (self.read_bytes > 0)
    did_writes = (self.write_bytes > 0)
    # we only count an open/close pair as an access
    # if there was actually I/O
    if not did_io:
      return

    agg.total_accesses += 1
    if did_reads and did_writes:
      agg.total_both_accesses += 1
    elif did_reads:
      agg.total_readonly_accesses += 1
    elif did_writes:
      agg.total_writeonly_accesses += 1

  def UpdateBytesAggregates(self):
    agg = self.exec_state.aggregates
    bytes = self.TotalIO()
    read_bytes = self.read_bytes
    jumps = self.jumps
    write_bytes = self.write_bytes
    agg.total_bytes += bytes
    agg.total_read_bytes += read_bytes
    agg.total_aio_read_bytes += self.aio_read_bytes
    agg.total_write_bytes += write_bytes
    if self.IsReadOnly():
      agg.total_readonly_bytes += read_bytes
    if self.IsWriteOnly():
      agg.total_writeonly_bytes += write_bytes
    # sequentiality stats
    if jumps == 0:
      agg.total_read_sequential100_bytes += read_bytes
      agg.total_write_sequential100_bytes += write_bytes
    if self.longest_run >= 0.95 * bytes:
      agg.total_read_sequential95_bytes += read_bytes
      agg.total_write_sequential95_bytes += write_bytes
    # preallocation stats
    agg.total_preallocation_bytes += self.TotalPreallocated()
    agg.total_extension_bytes += self.extension_bytes
    agg.total_preallocation_bytes_pwrite += self.pwrite_preallocated_bytes
    agg.preallocated_bytes_unnecessary += self.preallocated_bytes_unnecessary

  def UpdateFsyncAggregates(self):
    if self.fsync_count == 0:
      return
    agg = self.exec_state.aggregates
    agg.fsync_count += self.fsync_count
    agg.fsync_bytes += self.fsync_bytes
    agg.fsync_10mb_count += self.fsync_10mb_count
    agg.fsync_1mb_count += self.fsync_1mb_count
    agg.fsync_64kb_count += self.fsync_64kb_count
    agg.fsync_4kb_count += self.fsync_4kb_count
    agg.fsync_1b_count += self.fsync_1b_count
    agg.fsync_0b_count += self.fsync_0b_count

  def UpdateFileAggregates(self):
    # we only count a file if I/O was performed
    if self.TotalIO() == 0:
      return 
    file_size = self.inode.size
    agg = self.exec_state.aggregates
    if file_size >= 10*1024*1024:
      agg.file_10mb_count += 1
      agg.file_10mb_bytes += file_size
    elif file_size >= 1024*1024:
      agg.file_1mb_count += 1
      agg.file_1mb_bytes += file_size
    elif file_size >= 64*1024:
      agg.file_64kb_count += 1
      agg.file_64kb_bytes += file_size
    elif file_size >= 4*1024:
      agg.file_4kb_count += 1
      agg.file_4kb_bytes += file_size
    else:
      agg.file_0b_count += 1
      agg.file_0b_bytes += file_size
    agg.file_size_sum += file_size

  def UpdateOtherListeners(self):
    for listener in self.exec_state.listeners:
      listener.Close(self)

class IgnoreFd(Fd):
  def __init__(self):
    super(IgnoreFd, self).__init__()
    self.ignore = True


class Thread(object):
  def __init__(self, tid):
    self.tid = tid
    self.Reset()

  def TotalIO(self):
    return self.read_bytes + self.write_bytes

  def Reset(self):
    self.read_bytes = 0
    self.write_bytes = 0
    self.fsync_count = 0


class Aggregates(object):
  def __init__(self):
    self.Reset()

  def Reset(self):
    # bytes aggregates
    self.total_bytes = 0
    self.total_read_bytes = 0
    self.total_aio_read_bytes = 0
    self.total_write_bytes = 0
    self.total_readonly_bytes = 0
    self.total_writeonly_bytes = 0
    self.total_read_sequential95_bytes = 0
    self.total_write_sequential95_bytes = 0
    self.total_read_sequential100_bytes = 0
    self.total_write_sequential100_bytes = 0
    self.total_preallocation_bytes_pwrite = 0
    self.total_preallocation_bytes = 0
    self.total_extension_bytes = 0
    self.preallocated_bytes_unnecessary = 0
    # accesses aggregates
    self.total_accesses = 0
    self.total_both_accesses = 0
    self.total_readonly_accesses = 0
    self.total_writeonly_accesses = 0
    # file type aggregates
    self.total_bytes_category_productivity = 0
    self.total_bytes_category_multimedia = 0
    self.total_bytes_category_extension_strings = 0
    self.total_bytes_category_plist = 0
    self.total_bytes_category_sqlite = 0
    self.total_accesses_category_productivity = 0
    self.total_accesses_category_multimedia = 0
    self.total_accesses_category_extension_strings = 0
    self.total_accesses_category_plist = 0
    self.total_accesses_category_sqlite = 0
    # durability aggregates
    self.fsync_count = 0
    self.fsync_bytes = 0
    self.fsync_10mb_count = 0
    self.fsync_1mb_count = 0
    self.fsync_64kb_count = 0
    self.fsync_4kb_count = 0
    self.fsync_1b_count = 0
    self.fsync_0b_count = 0
    self.fsync_bytes_fsflushfork = 0
    self.fsync_bytes_preferences = 0
    self.fsync_bytes_sfarchiving = 0
    self.fsync_bytes_sqlite = 0
    self.fsync_bytes_writetofile_atomically = 0
    # atomicity aggregates
    self.rename_count = 0
    self.rename_bytes_diff_dir = 0
    self.rename_bytes_same_dir = 0
    self.exchangedata_count = 0
    self.exchangedata_bytes_diff_dir = 0
    self.exchangedata_bytes_same_dir = 0
    self.rename_count_cause_carbon_fsrenameunicode = 0
    self.rename_count_cause_multiple = 0
    self.rename_count_cause_nsdata_writetofile = 0
    self.rename_count_cause_nsfilemanager_movepath = 0
    self.rename_count_cause_preferences_synchronize = 0
    self.rename_count_cause_unknown = 0
    # file aggregates
    self.file_10mb_count = 0
    self.file_1mb_count = 0
    self.file_64kb_count = 0
    self.file_4kb_count = 0
    self.file_0b_count = 0
    self.file_10mb_bytes = 0
    self.file_1mb_bytes = 0
    self.file_64kb_bytes = 0
    self.file_4kb_bytes = 0
    self.file_0b_bytes = 0
    self.file_size_sum = 0
    # thread aggregates
    self.io_thread_count = 0
    self.threads_read_count = 0
    self.threads_write_count = 0
    self.threads_both_count = 0
    self.threads_both_and_fsync_count = 0
    self.threads_read_and_fsync_count = 0
    self.threads_write_and_fsync_count = 0
    self.thread1_io = 0
    self.thread2_io = 0
    self.thread3_io = 0
    self.thread4_io = 0
    self.thread5_io = 0
    self.thread6_io = 0
    self.thread7_io = 0
    self.thread8_io = 0
    self.thread9_io = 0
    self.thread10_io = 0
    self.thread11_io = 0
    self.thread12_io = 0
    self.thread13_io = 0
    self.thread14_io = 0
    self.thread15_io = 0
    self.thread16_io = 0
    self.thread17_io = 0
    self.thread18_io = 0
    self.thread19_io = 0
    self.thread20_io = 0
    self.thread21_io = 0
    self.thread22_io = 0
    self.thread23_io = 0
    self.thread24_io = 0
    self.thread25_io = 0
    self.thread26_io = 0
    self.thread27_io = 0
    self.thread28_io = 0
    self.thread29_io = 0
    self.thread30_io = 0
    # cpu aggregates
    self.cpu_seconds = 0  # this an upper bound
