import execution_state

class Event(object):
  name_map = None

  def __init__(self, args):
    self.a_time = args[0]
    self.a_tid = args[1]
    self.a_probe_type = args[2]
    self.a_syscall = args[3]

    # determines whether the event could return a file
    # descriptor.  In order to avoid tricky interleaving
    # issues, it is necessary that we process syscalls
    # that return fds at the return time.  It is also
    # necessary that we process the close syscall at the
    # entry time.  We could process the other syscalls
    # at either entry or return time, but we choose to
    # process them at entry time.
    self.generates_fd = False

  def __str__(self):
    fd = None
    if hasattr(self, 'a_fildes'):
      fd = self.a_fildes
    elif self.a_syscall.endswith('open'):
      fd = self.a_ret
    if fd is not None:
      return '%s %s, fd=%s' % (self.a_probe_type, self.a_syscall, fd)
    return '%s %s' % (self.a_probe_type, self.a_syscall)

  def execute(self, exec_state):
    pass

  @staticmethod
  def PopulateNameMap(args=[]):
    Event.name_map = {
      'open': Event_open,
      'shm_open': Event_shm_open,
      'close': Event_close,
      'dup': Event_dup,
      'kqueue': Event_kqueue,
      'socket': Event_socket,
      'socketpair': Event_socketpair,
      'read': Event_read,
      'write': Event_write,
      'pread': Event_pread,
      'pwrite': Event_pwrite,
      'aio_read': Event_aio_read,
      'aio_error': Event_aio_error,
      'aio_return': Event_aio_return,
      'aio_suspend': Event_aio_suspend,
      'lseek': Event_lseek,
      'fsync': Event_fsync,
      'fcntl': Event_fcntl,
      'ftruncate': Event_ftruncate,
      'link': Event_link,
      'symlink': Event_symlink,
      'rmdir': Event_rmdir,
      'unlink': Event_unlink,
      'readlink': Event_readlink,
      'rename': Event_rename,
      'exchangedata': Event_exchangedata,
      'mkdir': Event_mkdir,
      'flock': Event_flock,
      'mmap': Event_mmap,
      'futimes': Event_futimes,
      'utimes': Event_utimes,
      'statfs': Event_statfs,
      'statfs64': Event_statfs64,
      'fstatfs64': Event_fstatfs64,
      'fstat': Event_fstat,
      'lstat': Event_lstat,
      'stat': Event_stat,
      'fstat64': Event_fstat64,
      'lstat64': Event_lstat64,
      'stat64': Event_stat64,
      'setattrlist': Event_setattrlist,
      'getattrlist': Event_getattrlist,
      'getxattr': Event_getxattr,
      'fgetxattr': Event_fgetxattr,
      'listxattr': Event_listxattr,
      'flistxattr': Event_flistxattr,
      'access': Event_access,
      'chown': Event_chown,
      'chmod': Event_chmod,
      'fsetxattr': Event_fsetxattr,
      'fsgetpath': Event_fsgetpath,
      'getdirentries64': Event_getdirentries64,
      'getdirentriesattr': Event_getdirentriesattr,
      'getfsstat64': Event_getfsstat64,
      'getdtablesize': Event_getdtablesize,
      'chdir': Event_chdir,
      'fchdir': Event_fchdir,
      'kill': Event_kill,
      'fchflags': Event_fchflags,
      'fchmod': Event_fchmod,
      'fchown': Event_fchown,
      'removexattr': Event_removexattr,
      'fremovexattr': Event_fremovexattr,
      'setxattr': Event_setxattr,
      'controller': Event_controller,
      'resolver': Event_resolver,
      'all_calls': Event_all_calls,
      'setter': Event_setter,
      'build_path': Event_build_path,
      'namei': Event_namei,
      'vn_create': Event_vn_create,
      'hfs_vnop_pagein': Event_hfs_vnop_pagein,
      'vnode_pagein': Event_vnode_pagein,
      'syscall_map': Event_syscall_map,
      'posix_spawn': Event_posix_spawn,
      'execve': Event_execve,
    }

  @staticmethod
  def Create(args=[]):
    if not Event.name_map:
      Event.PopulateNameMap()
    event_class = Event.name_map[args[3]]
    return event_class(args)

class Event_open(Event):
  TYPE_FILE = 1
  TYPE_DIR = 2

  def __init__(self, args):
    super(Event_open, self).__init__(args)
    self.a_path = args[4]
    self.a_oflag = args[5]
    self.a_flags = args[6]
    self.a_type = args[7]
    self.a_inode = args[8]
    self.a_file_table = args[9]
    self.a_stack = args[10]
    self.a_ret = args[11]
    self.generates_fd = True

  def execute(self, exec_state):
    if self.a_ret < 0:
      return

    fd_object = None
    if self.a_type == Event_open.TYPE_FILE:
      flags = self.a_flags.split(' ')
      exec_state.OpenFile(timestamp = self.a_time,
                          fd_number = self.a_ret,
                          inode_number = self.a_inode,
                          path = self.a_path,
                          flags = flags)
    else:
      exec_state.AddDummyFd(self.a_ret)


class Event_shm_open(Event):
  def __init__(self, args):
    super(Event_shm_open, self).__init__(args)
    self.a_name = args[4]
    self.a_oflag = args[5]
    self.a_ret = args[6]
    self.generates_fd = True

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    exec_state.AddDummyFd(self.a_ret)


class Event_close(Event):
  def __init__(self, args):
    super(Event_close, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    exec_state.CloseFd(self.a_fildes)


class Event_dup(Event):
  def __init__(self, args):
    super(Event_dup, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    exec_state.DupFd(self.a_fildes, self.a_ret)


class Event_kqueue(Event):
  def __init__(self, args):
    super(Event_kqueue, self).__init__(args)
    self.a_ret = args[4]
    self.generates_fd = True

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    exec_state.AddDummyFd(self.a_ret)

class Event_socket(Event):
  def __init__(self, args):
    super(Event_socket, self).__init__(args)
    self.a_domain = args[4]
    self.a_type = args[5]
    self.a_protocol = args[6]
    self.a_ret = args[7]
    self.generates_fd = True

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    exec_state.AddDummyFd(self.a_ret)


class Event_socketpair(Event):
  def __init__(self, args):
    super(Event_socketpair, self).__init__(args)
    self.a_domain = args[4]
    self.a_type = args[5]
    self.a_protocol = args[6]
    self.a_fd1 = args[7]
    self.a_fd2 = args[8]
    self.a_ret = args[9]
    self.generates_fd = True

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    exec_state.AddDummyFd(self.a_fd1)
    exec_state.AddDummyFd(self.a_fd2)


class Event_read(Event):
  def __init__(self, args):
    super(Event_read, self).__init__(args)
    self.a_fildes = args[4]
    self.a_nbyte = args[5]
    self.a_offset = args[6]
    self.a_stack = args[7]
    self.a_ret = args[8]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    fd = exec_state.GetFd(self.a_fildes)
    if fd.ignore:
      return
    fd.Read(tid=self.a_tid, stack=self.a_stack, bytes=self.a_ret)

class Event_write(Event):
  def __init__(self, args):
    super(Event_write, self).__init__(args)
    self.a_fildes = args[4]
    self.a_nbyte = args[5]
    self.a_offset = args[6]
    self.a_stack = args[7]
    self.a_ret = args[8]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    fd = exec_state.GetFd(self.a_fildes)
    if fd.ignore:
      return
    fd.Write(tid=self.a_tid, stack=self.a_stack, bytes=self.a_ret)


class Event_pread(Event):
  def __init__(self, args):
    super(Event_pread, self).__init__(args)
    self.a_fildes = args[4]
    self.a_nbyte = args[5]
    self.a_offset = args[6]
    self.a_stack = args[7]
    self.a_ret = args[8]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    fd = exec_state.GetFd(self.a_fildes)
    if fd.ignore:
      return
    fd.Read(tid=self.a_tid, stack=self.a_stack, bytes=self.a_ret, offset=self.a_offset)


class Event_pwrite(Event):
  def __init__(self, args):
    super(Event_pwrite, self).__init__(args)
    self.a_fildes = args[4]
    self.a_nbyte = args[5]
    self.a_offset = args[6]
    self.a_stack = args[7]
    self.a_ret = args[8]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    fd = exec_state.GetFd(self.a_fildes)
    if fd.ignore:
      return
    fd.Write(tid=self.a_tid, stack=self.a_stack, bytes=self.a_ret, offset=self.a_offset)



class Event_aio_read(Event):
  def __init__(self, args):
    super(Event_aio_read, self).__init__(args)
    self.a_aiocbp = args[4]
    self.a_fildes = args[5]
    self.a_offset = args[6]
    self.a_size = args[7]
    self.a_retsize = args[8]
    self.a_stack = args[9]
    self.a_ret = args[10]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    fd = exec_state.GetFd(self.a_fildes)
    if fd.ignore:
      return
    fd.Read(tid=self.a_tid, stack=self.a_stack, bytes=self.a_retsize, offset=self.a_offset, aio=True)


class Event_aio_error(Event):
  def __init__(self, args):
    super(Event_aio_error, self).__init__(args)
    self.a_aiocbp = args[4]
    self.a_ret = args[5]


class Event_aio_return(Event):
  def __init__(self, args):
    super(Event_aio_return, self).__init__(args)
    self.a_aiocbp = args[4]
    self.a_ret = args[5]


class Event_aio_suspend(Event):
  def __init__(self, args):
    super(Event_aio_suspend, self).__init__(args)
    self.a_nent = args[4]
    self.a_ret = args[5]


class Event_lseek(Event):
  def __init__(self, args):
    super(Event_lseek, self).__init__(args)
    self.a_fildes = args[4]
    self.a_offset = args[5]
    self.a_whence = args[6]
    self.a_ret = args[7]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    fd = exec_state.GetFd(self.a_fildes)
    if fd.ignore:
      return
    fd.Seek(offset=self.a_ret)


class Event_fsync(Event):
  def __init__(self, args):
    super(Event_fsync, self).__init__(args)
    self.a_fildes = args[4]
    self.a_stack = args[5]
    self.a_ret = args[6]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    fd = exec_state.GetFd(self.a_fildes)
    if fd.ignore:
      return
    fd.Fsync(self.a_tid, self.a_stack)

class Event_fcntl(Event):
  def __init__(self, args):
    super(Event_fcntl, self).__init__(args)
    self.a_fildes = args[4]
    self.a_cmd = args[5]
    self.a_arg = args[6]
    self.a_cmd_name = args[7]
    self.a_stack = args[8]
    self.a_ret = args[9]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return

    if self.a_cmd_name == 'F_DUPFD':
      exec_state.DupFd(self.a_fildes, self.a_ret)
    elif self.a_cmd_name == 'F_FULLFSYNC':
      fd = exec_state.GetFd(self.a_fildes)
      if fd.ignore:
        return
      fd.Fsync(self.a_tid, self.a_stack)


class Event_ftruncate(Event):
  def __init__(self, args):
    super(Event_ftruncate, self).__init__(args)
    self.a_fildes = args[4]
    self.a_length = args[5]
    self.a_stack = args[6]
    self.a_ret = args[7]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    fd = exec_state.GetFd(self.a_fildes)
    if fd.ignore:
      return
    fd.Ftruncate(self.a_length)


class Event_link(Event):
  def __init__(self, args):
    super(Event_link, self).__init__(args)
    self.a_path1 = args[4]
    self.a_path2 = args[5]
    self.a_ret = args[6]


class Event_symlink(Event):
  def __init__(self, args):
    super(Event_symlink, self).__init__(args)
    self.a_path1 = args[4]
    self.a_path2 = args[5]
    self.a_ret = args[6]


class Event_rmdir(Event):
  def __init__(self, args):
    super(Event_rmdir, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_unlink(Event):
  def __init__(self, args):
    super(Event_unlink, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_readlink(Event):
  def __init__(self, args):
    super(Event_readlink, self).__init__(args)
    self.a_path = args[4]
    self.a_bufsize = args[5]
    self.a_target = args[6]
    self.a_ret = args[7]


class Event_rename(Event):
  def __init__(self, args):
    super(Event_rename, self).__init__(args)
    self.a_old_path = args[4]
    self.a_new_path = args[5]
    self.a_inode = args[6]
    self.a_type = args[7]
    self.a_from_dir_inode = args[8]
    self.a_to_dir_inode = args[9]
    self.a_stack = args[10]
    self.a_ret = args[11]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    if self.a_type != Event_open.TYPE_FILE:
      # TODO(tylerharter) - perhaps it would be interesting
      # to get rename stats on directories as well
      return
    inode = exec_state.GetInode(self.a_inode)
    inode.Rename(self.a_stack,
                 self.a_new_path,
                 self.a_from_dir_inode, 
                 self.a_to_dir_inode)


class Event_exchangedata(Event):
  def __init__(self, args):
    super(Event_exchangedata, self).__init__(args)
    self.a_path1 = args[4]
    self.a_path2 = args[5]
    self.a_inode1 = args[6]
    self.a_inode2 = args[7]
    self.a_ret = args[8]

  def execute(self, exec_state):
    if self.a_ret < 0:
      return
    path1_parts = self.a_path1.split('/')
    path1_parts.pop()
    path2_parts = self.a_path2.split('/')
    path2_parts.pop()
    same_dir = False
    if '/'.join(path1_parts) == '/'.join(path2_parts):
      same_dir = True

    if not same_dir:
      # TODO: if we set same_dir to True, then the files
      # are definitely in the same directory.  If, however,
      # we set same_dir to False, it is still possible that
      # the files are in the same directory, but that symlinks
      # allowed different paths to be used for the diretory.
      # If a trace ever has exchangedata's that appear to
      # be across directories, this code will need to be
      # revisited, but for our original iBench traces, this
      # is never the case.
      msg = ('We have not written code to address cases '
             'where exchangedata is called on two paths '
             'with different prefixes because we never '
             'observed this in our original traces')
      raise Exception(msg)

    inode1 = exec_state.GetInode(self.a_inode1)
    inode2 = exec_state.GetInode(self.a_inode2)
    inode1.ExchangeData(inode2, self.a_path2, same_dir)
    inode2.ExchangeData(inode1, self.a_path1, same_dir)


class Event_mkdir(Event):
  def __init__(self, args):
    super(Event_mkdir, self).__init__(args)
    self.a_path = args[4]
    self.a_mode = args[5]
    self.a_ret = args[6]


class Event_flock(Event):
  def __init__(self, args):
    super(Event_flock, self).__init__(args)
    self.a_fd = args[4]
    self.a_operation = args[5]
    self.a_stack = args[6]
    self.a_ret = args[7]


class Event_mmap(Event):
  def __init__(self, args):
    super(Event_mmap, self).__init__(args)
    self.a_addr = args[4]
    self.a_len = args[5]
    self.a_prot = args[6]
    self.a_flags = args[7]
    self.a_fd = args[8]
    self.a_offset = args[9]
    self.a_protstr = args[10]
    self.a_flagsstr = args[11]
    self.a_ret = args[12]


class Event_futimes(Event):
  def __init__(self, args):
    super(Event_futimes, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]


class Event_utimes(Event):
  def __init__(self, args):
    super(Event_utimes, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]


class Event_statfs(Event):
  def __init__(self, args):
    super(Event_statfs, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_statfs64(Event):
  def __init__(self, args):
    super(Event_statfs64, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_fstatfs64(Event):
  def __init__(self, args):
    super(Event_fstatfs64, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]


class Event_fstat(Event):
  def __init__(self, args):
    super(Event_fstat, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]


class Event_lstat(Event):
  def __init__(self, args):
    super(Event_lstat, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_stat(Event):
  def __init__(self, args):
    super(Event_stat, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_fstat64(Event):
  def __init__(self, args):
    super(Event_fstat64, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]


class Event_lstat64(Event):
  def __init__(self, args):
    super(Event_lstat64, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_stat64(Event):
  def __init__(self, args):
    super(Event_stat64, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_setattrlist(Event):
  def __init__(self, args):
    super(Event_setattrlist, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_getattrlist(Event):
  def __init__(self, args):
    super(Event_getattrlist, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_getxattr(Event):
  def __init__(self, args):
    super(Event_getxattr, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_fgetxattr(Event):
  def __init__(self, args):
    super(Event_fgetxattr, self).__init__(args)
    self.a_fd = args[4]
    self.a_ret = args[5]


class Event_listxattr(Event):
  def __init__(self, args):
    super(Event_listxattr, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_flistxattr(Event):
  def __init__(self, args):
    super(Event_flistxattr, self).__init__(args)
    self.a_fd = args[4]
    self.a_ret = args[5]


class Event_access(Event):
  def __init__(self, args):
    super(Event_access, self).__init__(args)
    self.a_path = args[4]
    self.a_amode = args[5]
    self.a_ret = args[6]


class Event_chown(Event):
  def __init__(self, args):
    super(Event_chown, self).__init__(args)
    self.a_path = args[4]
    self.a_owner = args[5]
    self.a_group = args[6]
    self.a_ret = args[7]


class Event_chmod(Event):
  def __init__(self, args):
    super(Event_chmod, self).__init__(args)
    self.a_path = args[4]
    self.a_mode = args[5]
    self.a_ret = args[6]


class Event_fsetxattr(Event):
  def __init__(self, args):
    super(Event_fsetxattr, self).__init__(args)
    self.a_fildes = args[4]
    self.a_name = args[5]
    self.a_ret = args[6]


class Event_fsgetpath(Event):
  def __init__(self, args):
    super(Event_fsgetpath, self).__init__(args)
    self.a_fsid = args[4]
    self.a_ret = args[5]


class Event_getdirentries64(Event):
  def __init__(self, args):
    super(Event_getdirentries64, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]


class Event_getdirentriesattr(Event):
  def __init__(self, args):
    super(Event_getdirentriesattr, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]


class Event_getfsstat64(Event):
  def __init__(self, args):
    super(Event_getfsstat64, self).__init__(args)
    self.a_ret = args[4]


class Event_getdtablesize(Event):
  def __init__(self, args):
    super(Event_getdtablesize, self).__init__(args)
    self.a_ret = args[4]


class Event_chdir(Event):
  def __init__(self, args):
    super(Event_chdir, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_fchdir(Event):
  def __init__(self, args):
    super(Event_fchdir, self).__init__(args)
    self.a_fildes = args[4]
    self.a_ret = args[5]


class Event_kill(Event):
  def __init__(self, args):
    super(Event_kill, self).__init__(args)
    self.a_pid = args[4]
    self.a_sig = args[5]
    self.a_ret = args[6]


class Event_fchflags(Event):
  def __init__(self, args):
    super(Event_fchflags, self).__init__(args)
    self.a_fd = args[4]
    self.a_flags = args[5]
    self.a_ret = args[6]


class Event_fchmod(Event):
  def __init__(self, args):
    super(Event_fchmod, self).__init__(args)
    self.a_fildes = args[4]
    self.a_mode = args[5]
    self.a_ret = args[6]


class Event_fchown(Event):
  def __init__(self, args):
    super(Event_fchown, self).__init__(args)
    self.a_fildes = args[4]
    self.a_owner = args[5]
    self.a_group = args[6]
    self.a_ret = args[7]


class Event_removexattr(Event):
  def __init__(self, args):
    super(Event_removexattr, self).__init__(args)
    self.a_path = args[4]
    self.a_name = args[5]
    self.a_ret = args[6]


class Event_fremovexattr(Event):
  def __init__(self, args):
    super(Event_fremovexattr, self).__init__(args)
    self.a_fd = args[4]
    self.a_name = args[5]
    self.a_ret = args[6]


class Event_setxattr(Event):
  def __init__(self, args):
    super(Event_setxattr, self).__init__(args)
    self.a_path = args[4]
    self.a_name = args[5]
    self.a_ret = args[6]


class Event_controller(Event):
  def __init__(self, args):
    super(Event_controller, self).__init__(args)
    self.a_msg = args[4]


class Event_resolver(Event):
  def __init__(self, args):
    super(Event_resolver, self).__init__(args)
    self.a_vol = args[4]
    self.a_resolved = args[5]
    self.a_original_path = args[6]
    self.a_comments = args[7]


class Event_all_calls(Event):
  def __init__(self, args):
    super(Event_all_calls, self).__init__(args)
    self.a_name = args[4]


class Event_setter(Event):
  def __init__(self, args):
    super(Event_setter, self).__init__(args)
    self.a_variable = args[4]
    self.a_value = args[5]


class Event_build_path(Event):
  def __init__(self, args):
    super(Event_build_path, self).__init__(args)
    self.a_inode = args[4]
    self.a_path = args[5]


class Event_namei(Event):
  def __init__(self, args):
    super(Event_namei, self).__init__(args)
    self.a_path = args[4]
    self.a_resolved = args[5]


class Event_vn_create(Event):
  def __init__(self, args):
    super(Event_vn_create, self).__init__(args)
    self.a_inode = args[4]


class Event_hfs_vnop_pagein(Event):
  def __init__(self, args):
    super(Event_hfs_vnop_pagein, self).__init__(args)
    self.a_offset = args[4]
    self.a_size = args[5]
    self.a_inode = args[6]
    self.a_name = args[7]
    self.a_actual_size = args[8]
    self.a_ret = args[9]


class Event_vnode_pagein(Event):
  def __init__(self, args):
    super(Event_vnode_pagein, self).__init__(args)
    self.a_offset = args[4]
    self.a_size = args[5]
    self.a_inode = args[6]
    self.a_name = args[7]
    self.a_actual_size = args[8]
    self.a_ret = args[9]


class Event_syscall_map(Event):
  def __init__(self, args):
    super(Event_syscall_map, self).__init__(args)
    self.a_map = args[4]


class Event_posix_spawn(Event):
  def __init__(self, args):
    super(Event_posix_spawn, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]


class Event_execve(Event):
  def __init__(self, args):
    super(Event_execve, self).__init__(args)
    self.a_path = args[4]
    self.a_ret = args[5]



