import os
import re
import tarfile
import execution_state
import trace_event


class AnalyzeException(Exception):
  def __init__(self, *args, **kwargs):
    super(AnalyzeException, self).__init__(*args, **kwargs)


class FileSystemSnapshot(object):
  def __init__(self, snapshot_file):
    self.file = snapshot_file

  def __iter__(self):
    return self

  def next(self):
    line = self.file.readline()
    while line:
      args = TraceReader.ParseArgs(line)
      if args[0] == 'file':
        _, inode_number, file_path, size = args
        inode = execution_state.Inode(inode_number, file_path, size)
        return inode
      line = self.file.readline()
    raise StopIteration()


class TraceReader:
  def __init__(self, app, path):
    self.trace_tar = tarfile.open(path, 'r:gz')
    # get event FD
    self.file = self.trace_tar.extractfile('sorted.txt')
    self.bytes_read = 0
    self.file_size = self.file.size
    # get snapshot FD
    self.snapshot_file = self.trace_tar.extractfile('init.txt')
    self.snapshot = FileSystemSnapshot(self.snapshot_file)
    # get CPU data from top
    top_file = self.trace_tar.extractfile('top.log')
    self.cpu_values = self.GetCPUUsage(app, top_file)
    top_file.close()
    # updated as we go
    self.last_timestamp = 0

  def Close(self):
    self.file.close()
    self.snapshot_file.close()
    self.trace_tar.close()

  def GetCPUUsage(self, app, top_file):
    output_start = 'COMMAND          %CPU'
    output_end = 'Processes: '
    cpu_values = []
    cpu_percent_section = False
    for line in top_file:
      if line.startswith(output_start):
         cpu_percent_section = True
      elif line.startswith(output_end):
         cpu_percent_section = False
      elif cpu_percent_section:
        if line.lower().startswith(app.lower() + ' '):
          m = re.search(r' ([^ ]+)', line)
          if not m:
            msg = ('Could not determine CPU from top entry '
                   '%s' % line)
            raise Exception(msg)
          value = float(m.group(1).strip())/100.0

          # sometimes top produces garbage values.  A comment in the code of
          # the Mac implementation says this is fine and that numbers will
          # stabalize in the next frame.  Thus, we ignore garbage values that
          # are not in a reasonable range.  This is very rare, and it seems that
          # the CPU utilization in the garbage frame just gets added to the
          # next frame, so this should produce reasonable if not perfect
          # results.
          if value >= 20 or value < 0:
            value = 0
          cpu_values.append(value)
    if not cpu_values:
      raise Exception('Could not find any CPU values '
                      'for app "%s"' % app)
    return cpu_values

  @staticmethod
  def ParseArgs(line):
    line = line.rstrip('\n')
    args = line.split(']; [')
    args[0] = args[0][1:]
    args[-1] = args[-1][:len(args[-1]) - 3]
    for i in range(len(args)):
      if re.match(r'\-?\d+$', args[i]):
        args[i] = int(args[i])
    return args
    
  def Percent(self):
    return int(100 * self.bytes_read / self.file_size)

  def __iter__(self):
    return self

  def next(self):
    line = self.file.readline()
    self.bytes_read += len(line)
    if line:
      args = TraceReader.ParseArgs(line)
      event = trace_event.Event.Create(args)

      if(event.a_probe_type != 'comment'):
        self.last_timestamp = event.a_time

      return event
    else:
      raise StopIteration()
